import codecs
import json
import os
from pathlib import Path
import shutil
from packaging import version

from lyrebird.mock.logger_helper import get_logger
from .version import IVERSION
from typing import List
from flask import Response, abort
from lyrebird.mock.console_helper import warning_msg, err_msg
import subprocess

"""
Config manager

Config file path: ~/.lyrebird/conf.json
Lyrebird will copy config template to ~/.lyrebird/conf.json, when lyrebird server start. 
"""

CURRENT_DIR = Path(__file__).parent

CONFIG_TEMPLATE_FILE = CURRENT_DIR/'templates'/'conf.json'
CACHE_TEMPLATE_FILE = CURRENT_DIR/'templates'/'cache.json'


class Config:

    def __init__(self, root_dir='~', name='lyrebird'):
        self.root = Path(root_dir, '.lyrebird').expanduser()
        self.default_conf_path = Path(self.root, 'conf.json').expanduser()
        self.custom_conf_dir = Path(self.root, 'conf').expanduser()
        self.cache_path = Path(self.root, 'cache.json').expanduser()
        self.pid_path = Path(self.root, f'{name}.pid').expanduser()
        self.tmp_dir = Path(self.root, 'tmp').expanduser()
        self.plugin_root = Path(self.root, 'plugins')

    def init(self):
        if not self.root.exists():
            self.root.mkdir()
        if not self.default_conf_path.exists():
            shutil.copyfile(CONFIG_TEMPLATE_FILE, self.default_conf_path)
        else:
            current_conf_version = self.load_default().get('version', '0.0.0')
            template_conf_version = self.load_template().get('version', '0.0.0')
            if version.parse(current_conf_version) < version.parse(template_conf_version):
                shutil.copyfile(CONFIG_TEMPLATE_FILE, self.default_conf_path)
        if not self.cache_path.exists():
            shutil.copyfile(CACHE_TEMPLATE_FILE, self.cache_path)
        if not self.custom_conf_dir.exists():
            self.custom_conf_dir.mkdir(parents=True)
        if not self.tmp_dir.exists():
            self.tmp_dir.mkdir(parents=True)

    def load(self, name):
        conf = self.load_default()
        if name:
            conf.update(self.load_custom(name))
        else:
            self.load_cache().get('custom_conf')
        return conf

    def load_template(self):
        return json.loads(codecs.open(CONFIG_TEMPLATE_FILE).read())

    def load_default(self):
        return json.loads(codecs.open(self.default_conf_path).read())

    def load_custom(self, name):
        if not name.endswith('.json'):
            name = name + '.json'
        custom_conf_path = self.custom_conf_dir / name
        if custom_conf_path.exists():
            return json.loads(codecs.open(self.custom_conf_dir/name).read())
        else:
            return json.loads(codecs.open(self.custom_conf_dir/'tmp.json').read())

    def save_pid(self):
        with codecs.open(self.pid_path, 'w', 'utf-8') as f:
            f.write(str(os.getpid()))

    def read_pid(self):
        with codecs.open(self.pid_path, 'r', 'utf-8') as f:
            return int(f.read())

    def remove_pid(self):
        if self.pid_path.exists():
            os.remove(self.pid_path)

    def load_tmp(self, name, **kwargs):
        tmp_conf_path = self.custom_conf_dir/'tmp.json'
        with codecs.open(tmp_conf_path, 'w', 'utf-8') as f:
            f.write(json.dumps(kwargs, indent=4, ensure_ascii=False))
        conf = self.load(name)
        conf.update(kwargs)
        return conf

    def load_cache(self):
        if self.cache_path.exists():
            with codecs.open(self.cache_path, 'r', 'utf-8') as f:
                return json.loads(f.read())
        else:
            raise ConfCacheNotFound

    def save_cache(self, cache):
        with codecs.open(self.cache_path, 'w', 'utf-8') as f:
            f.write(json.dumps(cache, indent=4, ensure_ascii=False))

    def download(self, uri):
        repo = self.tmp_dir/'repo'
        if repo.exists():
            shutil.rmtree(repo)
        p = subprocess.run(f"git clone {uri} {repo}", shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if p.returncode == 0:
            print(p.stdout.decode())
        else:
            # raise GitCloneFailed(p.stderr.decode())
            get_logger().critical(
                msg=f'下载配置文件错误,但不影响Overbridge使用,可以手动import相应插件base \n download conf data by git failed:{p.stderr.decode()}')
            return
        for subfile in repo.iterdir():
            if not subfile.is_dir():
                continue
            if subfile.name.startswith('.'):
                continue
            if subfile.name == 'conf':
                for conf_file in subfile.iterdir():
                    shutil.copy(conf_file, self.custom_conf_dir/conf_file.name)
            else:
                try:
                    dst = os.path.join(self.plugin_root, subfile.name)
                    command = f'cp -f -R {subfile}/* {dst}'
                    if not os.path.exists(dst):
                        os.makedirs(dst)
                    subprocess.run(command, shell=True)
                except Exception as e:
                    print(e)


class ConfCacheNotFound(Exception):
    pass


class GitCloneFailed(Exception):
    pass



from colorama import Fore, Style

from . import context


def tag_green(msg):
    if not msg:
        return ''
    return f'{Fore.GREEN}[{msg}]{Style.RESET_ALL}'


def tag_yello(msg):
    if not msg:
        return ''
    return f'{Fore.YELLOW}[{msg}]{Style.RESET_ALL}'


def warning_msg(*msg, check_verbose=False):
    print(f'{Fore.YELLOW}lyrebird', *msg, Style.RESET_ALL)


def mock_msg(*msg, check_verbose=False):
    print(f'{Fore.GREEN}lyrebird{Style.RESET_ALL}', *msg)


def err_msg(*msg, check_verbose=False):
    print(f'{Fore.RED}lyrebird Error{Style.RESET_ALL}', *msg)


def url_color(line):
    return f'{Fore.BLUE}{line}{Style.RESET_ALL}'


def statistic_msg(*msg, check_verbose=False):
    print(f'{Fore.RED}[STATISTIC_MSG]{Style.RESET_ALL}', *msg)



import logging
import colorama
import copy
from logging.config import fileConfig
from .logger_config import LoggingConfig

def _print_error(msg):
    print('\n!!!!!!!!!!!!!!!!!!!!!!!!\n!!!    %s\n!!!!!!!!!!!!!!!!!!!!!!!!\n' % msg)


LOG_COLORS = {
    logging.ERROR: [
        colorama.Fore.RED,
        colorama.Style.BRIGHT,
        colorama.Back.RESET
    ],
    logging.WARNING: [
        colorama.Fore.YELLOW,
        colorama.Style.NORMAL,
        colorama.Back.RESET
    ],
    logging.INFO: [
        colorama.Fore.WHITE,
        colorama.Style.NORMAL,
        colorama.Back.RESET,
    ],
    logging.CRITICAL: [
        colorama.Fore.WHITE,
        colorama.Style.BRIGHT,
        colorama.Back.RED,
    ],
    logging.DEBUG: [
        colorama.Fore.GREEN,
        colorama.Style.NORMAL,
        colorama.Back.RESET
    ]
}


def init_logger_settings(*, verbose=False):
    logging_config = LoggingConfig(verbose)
    try:
        logging.config.fileConfig(
            logging_config.get_user_logging_config(),
            disable_existing_loggers=False)
    except SyntaxError:
        '''
        若获取用户 Logging 配置文件，更新的过程中出现无法解析异常
        则 使用默认的配置加载
        '''
        logging.config.fileConfig(
            logging_config.use_default_logging_config(),
            disable_existing_loggers=False)

    socketio = logging.getLogger('socketio')
    socketio.setLevel(logging.ERROR)
    socketio.disabled = True

    engineio = logging.getLogger('engineio')
    engineio.setLevel(logging.ERROR)
    engineio.disabled = True

    mock = logging.getLogger('mock')
    mock.setLevel(logging.DEBUG)
    mock.addHandler(logging.StreamHandler())
    mock.disabled = True

    werkzeug = logging.getLogger('werkzeug')
    werkzeug.setLevel(logging.ERROR)
    werkzeug.disabled = True


class ColorFormatter(logging.Formatter):
    def format(self, record, *args, **kwargs):
        new_record = copy.copy(record)
        new_record.levelname = "{fore}{style}{back}{level}{end}".format(
            level=new_record.levelname,
            fore=LOG_COLORS[new_record.levelno][0],
            style=LOG_COLORS[new_record.levelno][1],
            back=LOG_COLORS[new_record.levelno][2],
            end=colorama.Style.RESET_ALL,
        )

        new_record.module = "{fore}{style}{module}{end}".format(
            module=new_record.module,
            fore=colorama.Fore.WHITE,
            style=colorama.Style.BRIGHT,
            end=colorama.Style.RESET_ALL
        )

        new_record.msg = "{fore}{style}{back}{msg}{end}".format(
            msg=new_record.msg,
            fore=LOG_COLORS[new_record.levelno][0],
            style=LOG_COLORS[new_record.levelno][1],
            back=LOG_COLORS[new_record.levelno][2],
            end=colorama.Style.RESET_ALL
        )

        return super(ColorFormatter, self).format(new_record)


def get_logger():
    """
    获取 logger 对象用于打印日志信息。
    e.g 获取 logger 对象
    _logger = get_logger()

    e.g 输出日志信息
    _logger.info('要输出的INFO日志信息')
    _logger.debug('要输出的DEBUG日志信息')

    :return: logger 对象
    """
    logger = logging.getLogger()
    return logger



import codecs
import configparser
from pathlib import Path


not_set_section = ['loggers', 'handlers', 'formatters', 'logger_root']


def set_logger_config(conf, *, logger_section, logger_options, logger_value):
    if logger_section in not_set_section:
        return
    if not conf.has_section(logger_section):
        conf.add_section(logger_section)
    conf.set(logger_section, logger_options, logger_value)


class LoggingConfig:

    USER_ROOT = Path('~', '.lyrebird').expanduser()
    CURRENT_PATH = Path(__file__).resolve().parent
    LOGGER_TMP_CONFIG = CURRENT_PATH / '../templates' / 'logging_config.ini'
    LOGFILE_PATH = USER_ROOT / 'log'
    LOGGING_CONFIG = 'logging_config.ini'
    USER_LOGGING_CONFIG = USER_ROOT / LOGGING_CONFIG

    def __init__(self, verbose):
        self.init_log_dir()
        self.verbose = verbose

    def init_log_dir(self):
        """
        初始化 ~/.lyrebird/log 目录

        TODO: 该操作需要考虑是否加入 filesystem 中管理，保持时序性。
        :return:
        """
        self.LOGFILE_PATH.mkdir(parents=True, exist_ok=True)

    def get_level(self):
        """
        启动参数带 v， 则 level 为 DEBUG
        启动参数不带 v，则 level 为 INFO

        :return: "DEBUG" or "INFO"
        """
        return "DEBUG" if self.verbose else "INFO"

    def init_logger_config(self):
        if self.USER_LOGGING_CONFIG.exists():
            self.update_logger_config()
        else:
            self.create_logger_config()

    def update_logger_config(self):
        user_conf = configparser.RawConfigParser()
        user_conf.read(self.USER_LOGGING_CONFIG)
        default_conf = configparser.RawConfigParser()
        default_conf.read(self.LOGGER_TMP_CONFIG)
        for section in user_conf.sections():
            for option in user_conf.options(section):
                set_logger_config(
                    conf=default_conf,
                    logger_section=section,
                    logger_options=option,
                    logger_value=user_conf.get(section, option)
                )
        # Update handler_stream - Level from start params <v>
        set_logger_config(default_conf, logger_section='handler_stream', logger_options='level', logger_value=self.get_level())
        with codecs.open(self.USER_LOGGING_CONFIG, 'w', 'utf-8') as f:
            default_conf.write(f)

    def create_logger_config(self):
        conf = configparser.RawConfigParser()
        conf.read(self.LOGGER_TMP_CONFIG)
        set_logger_config(
            conf,
            logger_section='handler_logFileHandler',
            logger_options='args',
            logger_value='("%s", "midnight", 1, 5, "utf-8")' % (self.USER_ROOT / 'log' / 'lyrebird.log')
        )
        # Update handler_stream - Level from start params <v>
        set_logger_config(conf, logger_section='handler_stream', logger_options='level', logger_value=self.get_level())
        with codecs.open(self.USER_LOGGING_CONFIG, 'w', 'utf-8') as f:
            conf.write(f)

    def get_user_logging_config(self):
        self.init_logger_config()
        return self.USER_ROOT / self.LOGGING_CONFIG

    def use_default_logging_config(self):
        self.create_logger_config()
        return self.USER_ROOT / self.LOGGING_CONFIG



from flask import jsonify

"""
Lyrebird context

"""

def make_ok_response(**kwargs):
    ok_resp = {
            "code": 1000,
            "message": "success"
        }
    ok_resp.update(kwargs)
    return jsonify(ok_resp)


def make_fail_response(msg):
    return jsonify(
        {
            "code": 3000,
            "message": msg
        }
    )


config = {}
server = {}
plugins = {}


def start_server():
    for name in server:
        server[name].start()



import os
import sys
from signal import SIGTERM
import signal
import socket
import fire
from lyrebird import config
from .mock.logger_helper import init_logger_settings
import webbrowser
import threading
from lyrebird.project_builder.builder import PluginProjectBuilder
from lyrebird.mock.mock_server import LyrebirdMockServer
from lyrebird.proxy.proxy_server import LyrebirdProxyServer
from lyrebird.event import EventServer
from lyrebird.background import BackgroundTaskServer
from lyrebird import application


"""
Lyrebird main entry

"""

class ManagerError(Exception):
    pass


class Server:
    """
    Lyrebird server class

    useage:
        import lyrebird
        
        server = lyrebird.Server()
        server.start()
    """

    def __init__(self):
        self.verbose = False
        self.record = False
        self.custom_conf_name = None
        self._conf = None
        self._pid = None
        self._mock_port = None
        self._proxy_port = None
        self._data_root_dir = None
        self.mock_server = None
        self.proxy_server = None
        self._conf_manager = None
        self.pid_file_name = None

    def v(self):
        """
        是否显示详细日志

        """
        self.verbose = True

    def mock_port(self, port):
        self._mock_port = port
        return self

    def proxy_port(self, port):
        self._proxy_port = port
        return self

    def data_root_dir(self, path):
        self._data_root_dir = path
        return self

    def set_cache_uri(self, uri):
        self._conf_manager = config.Config(name=self.pid_file_name)
        self._conf_manager.init()
        cache = self._conf_manager.load_cache()
        cache['user_data_uri'] = uri
        self._conf_manager.save_cache(cache)

    def init_conf(self):
        self._conf_manager = config.Config(name=self.pid_file_name)
        self._conf_manager.init()
        self._conf_manager.save_pid()

        custom_args = {}

        if self._mock_port:
            custom_args['mock.port'] = self._mock_port
        if self._proxy_port:
            custom_args['proxy.port'] = self._proxy_port
        if self._data_root_dir:
            custom_args['mock.data'] = self._data_root_dir

        cache = self._conf_manager.load_cache()

        if self.custom_conf_name == None:
            self.custom_conf_name = cache.get('custom_conf')
        else:
            cache['custom_conf'] = self.custom_conf_name
            self._conf_manager.save_cache(cache)

        if cache.get('user_data_uri'):
            self._conf_manager.download(cache.get('user_data_uri'))

        self._conf = self._conf_manager.load_tmp(self.custom_conf_name, **custom_args)
        self._conf['ip'] = get_ip()

        application.config = self._conf

    def start(self):
        """
        启动proxy&mock

        """
        init_logger_settings(verbose=self.verbose)
        self.init_conf()
        
        application.server['event'] = EventServer()
        application.server['background'] = BackgroundTaskServer()
        application.server['proxy'] = LyrebirdProxyServer()   
        application.server['mock'] = LyrebirdMockServer()

        application.start_server()



    def stop(self):
        if self.mock_server:
            self.mock_server.stop()


class CommandLine:
    """
    命令行入口
    
    * lyrebird
    以缺省参数启动lyrebird

    * lyrebird start
    以缺省参数启动lyrebird

    * lyrebird v start
    以输出详细日志模式启动lyrebird

    * lyrebird no-browser start
    启动lyrebird不默认打开浏览器

    * lyrebird start --mock 9090 --proxy 4272 --data .
    指定参数启动lyrebird
    参数：
        --mock 默认9090 ， mock服务及前端端口
        --proxy 默认4272， 代理服务端口
        --data 默认./data, mock数据根目录
        --name 默认lyrebird，服务别名（用于通过别名停止指定的lyrebird服务）

    * lyrebird stop
    停止lyrebird

    * lyrebird stop --name foo
    停止别名为foo的lyrebird
    """
    def __init__(self):
        self._open_browser = True
        self.server = Server()

    def v(self):
        self.server.v()
        return self

    def no_browser(self):
        self._open_browser = False
        return self

    def start(self, mock=None, proxy=None, data=None, name='lyrebird', config=None):
        if mock:
            self.server.mock_port(mock)
        if proxy:
            self.server.proxy_port(proxy)
        if data:
            self.server.data_root_dir(data)
        if name:
            self.server.pid_file_name = name
        if config:
            self.server.custom_conf_name = config

        # open browser
        def open_browser():
            if self._open_browser:
                webbrowser.open('http://localhost:' + str(self.server._conf.get('mock.port')))

        def signal_handler(signum, frame):
            print('\n!!!Ctrl-C pressed. Lyrebird stop!!!')
            self.server.stop()
            threading.Event().set()
            os._exit(1)

        signal.signal(signal.SIGINT, signal_handler)
        signal.signal(signal.SIGTERM, signal_handler)

        self.server.start()
        open_browser()

    def stop(self, name='lyrebird'):
        try:
            os.kill(config.Config(name=name).read_pid(), SIGTERM)
        except Exception:
            # 未找到进程时不处理
            pass
        config.Config(name=name).remove_pid()

    def resource(self, uri=None, branch='master'):
        if uri:
            uri = '%s --branch %s' % (uri, branch)
            self.server.set_cache_uri(uri)
        else:
            print('\nResource uri is not set!')


def debug():
    CommandLine().v().start()


def run():
    if len(sys.argv) == 1:
        fire.Fire(CommandLine, 'v start')
    else:
        fire.Fire(CommandLine)


def plugin():
    fire.Fire(PluginProjectBuilder)


def get_ip():
    """
    获取当前设备在网络中的ip地址

    :return: IP地址字符串
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(('meituan.com', 80))
    return s.getsockname()[0]